﻿using System;
using System.Collections.Generic;
using System.ComponentModel.DataAnnotations;
using System.Linq;
using IMSES_Template.DataAccess;
using IMSES_Template.ImsEnterpriseServicePort;
using IMSES_Template.Model;

namespace IMSES_Template.Utils
{
    public static class ConvertObjects
    {

        #region Constants

        private const string FieldString = "String";
        private const string FieldBoolean = "Boolean";

        private const string FirstName = "First";
        private const string LastName = "Last";

        private const string Custom0 = "customstring0";
        private const string Custom1 = "customstring1";
        private const string Custom2 = "customstring2";
        private const string Custom3 = "customstring3";
        private const string Custom4 = "customstring4";
        private const string EmailServerEmail = "emailserver/email";
        private const string EmailServerServer = "emailserver/server";
        private const string EmailServerPort = "emailserver/port";
        private const string EmailServerType = "emailserver/type";
        private const string EmailServerUserid = "emailserver/userid";
        private const string EmailServerPassword = "emailserver/userid/@password";

        private const string VisibleSearch = "visibleinsearch";
        private const string Scheme = "ItslearningOrganisationTypes";
        private const string LegalId = "legalid";

        private const string BoolTrue = "True";
        private const string BoolFalse = "False";

        private const string Course = "course";
        private const string CourseCode = "course/code";
        private const string CourseCredits = "course/credits";

        private const string Meta = "meta-";
        private const string HomeOrganisation = "HomeOrganisation";

        private const string HierarchyLevel = "-1";
        private const string SiteLevel = "0";
        private const string SchoolLevel = "1";
        private const string CourseLevel = "2";
        private const string CourseGroupLevel = "3";

        #endregion


        #region User

        /// <summary>
        /// Creates an Itslearning user based on the data from an ImsEs Person.
        /// </summary>
        /// <param name="imsEsPerson">The ImsEs person.</param>
        /// <param name="syncKey">The sync key of the user.</param>
        /// <returns>An Itslearning user.</returns>
        public static User ToItsUser(this person imsEsPerson, string syncKey)
        {
            // In case the read failed, return null.
            if (imsEsPerson == null) { return null; }

            var profile = imsEsPerson.institutionRole[0].institutionRoleType.ToProfile();
            var newUser = new User(syncKey)
                {
                    Profile = profile,
                    Email = imsEsPerson.email,

                    // Create the empty Objects, to be filled later:
                    CustomStrings = new CustomStrings
                        {
                            CustomString1 = null,
                            CustomString2 = null,
                            CustomString3 = null,
                            CustomString4 = null,
                            CustomString5 = null,
                        },
                    EmailServer = new EmailServer
                        {
                            Email = null,
                            Server = null,
                            Port = null,
                            Type = null,
                            UserId = null,
                            Password = null,
                        },

                    Children = new List<string>(),
                    MetaData = new List<MetaData>(),
                };

            // Read out the difficult data from the ImsEs person.
            if (imsEsPerson.name != null && imsEsPerson.name.partName != null)
            {
                foreach (var name in imsEsPerson.name.partName)
                {
                    var nameType = name.namePartType.ToNameType();

                    switch (nameType)
                    {
                        case NameFieldType.First:
                            newUser.FirstName = name.namePartValue;
                            break;
                        case NameFieldType.Last:
                            newUser.LastName = name.namePartValue;
                            break;
                    }
                }
            }

            if (imsEsPerson.userId != null)
            {
                newUser.UserName = imsEsPerson.userId.userIdValue;
                newUser.Password = imsEsPerson.userId.passWord;
            }

            if (imsEsPerson.address != null)
            {
                newUser.Address = new Address
                {
                    AddressLine1 = imsEsPerson.address.street[0],
                    AddressLine2 = imsEsPerson.address.extadd,
                    PostalCode = imsEsPerson.address.postcode,
                    City = imsEsPerson.address.locality
                };
            }

            if (imsEsPerson.demographics != null)
            {
                newUser.BirthDate = imsEsPerson.demographics.bday;
            }
            
            if (imsEsPerson.tel != null)
            {
                foreach (var telephone in imsEsPerson.tel)
                {
                    switch (telephone.telType)
                    {
                        case TelDTypeTelType.Voice:
                            newUser.Phone = telephone.telValue;
                            break;
                        case TelDTypeTelType.Mobile:
                            newUser.CellPhone = telephone.telValue;
                            break;
                    }
                }
            }

            if (imsEsPerson.extension!= null) 
            {
                if (imsEsPerson.extension.extensionField != null)
                {
                    foreach (var extension in imsEsPerson.extension.extensionField)
                    {
                        var extensionField = extension.fieldName.ToExtensionType();

                        switch (extensionField)
                        {
                            case ExtensionFieldType.Nationalidentitynumber:
                                newUser.CustomStrings.CustomString1 = extension.fieldValue;
                                break;
                            case ExtensionFieldType.Customstring0:
                                newUser.CustomStrings.CustomString1 = extension.fieldValue;
                                break;
                            case ExtensionFieldType.Customstring1:
                                newUser.CustomStrings.CustomString2 = extension.fieldValue;
                                break;
                            case ExtensionFieldType.Customstring2:
                                newUser.CustomStrings.CustomString3 = extension.fieldValue;
                                break;
                            case ExtensionFieldType.Customstring3:
                                newUser.CustomStrings.CustomString4 = extension.fieldValue;
                                break;
                            case ExtensionFieldType.Customstring4:
                                newUser.CustomStrings.CustomString5 = extension.fieldValue;
                                break;

                            case ExtensionFieldType.EmailserverEmail:
                                newUser.EmailServer.Email = extension.fieldValue;
                                break;
                            case ExtensionFieldType.EmailserverServer:
                                newUser.EmailServer.Server = extension.fieldValue;
                                break;
                            case ExtensionFieldType.EmailserverPort:
                                newUser.EmailServer.Port = extension.fieldValue;
                                break;
                            case ExtensionFieldType.EmailserverType:
                                newUser.EmailServer.Type = extension.fieldValue;
                                break;
                            case ExtensionFieldType.EmailserverUserid:
                                newUser.EmailServer.UserId = extension.fieldValue;
                                break;
                            case ExtensionFieldType.EmailserverUseridPassword:
                                newUser.EmailServer.Password = extension.fieldValue;
                                break;

                            case ExtensionFieldType.MetaData:
                                newUser.MetaData.Add(new MetaData
                                    {
                                        Type = extension.fieldName,
                                        Value = extension.fieldValue
                                    });
                                break;
                        }
                    }
                }

                if (imsEsPerson.extension.relationship != null)
                {
                    newUser.Children = imsEsPerson.extension.relationship
                                            .Where(x => x.relation == RelationshipDTypeRelation.Child)
                                            .Select(x => x.sourceId.identifier)
                                            .ToList();
                }
            }

            return newUser;
        }


        /// <summary>
        /// Converts a PersonIdPair array to a list of Itslearning Users.
        /// </summary>
        /// <param name="personPairList">The array with PersonPairs.</param>
        /// <returns>An list with Users.</returns>
        public static List<User> ToItsUserList(this PersonIdPairDType[] personPairList)
        {
            return personPairList.Select(x => x.person.ToItsUser(x.sourcedId.identifier)).ToList();
        }


        /// <summary>
        /// Converts the list with persons into a list with Users, based on the synckeys and the response 
        /// (to check if all users were actually received from Itslearning.)
        /// </summary>
        /// <param name="personList">List with persons.</param>
        /// <param name="syncKeyList">List with synckeys.</param>
        /// <param name="response">Statusinfo to check if all persons are actually retrieved.</param>
        /// <returns>The list with Users.</returns>
        public static List<User> ToItsUserList(this person[] personList, List<string> syncKeyList, syncResponseHeaderInfo response)
        {
            // For Batch functions, if there is only 1 object left, the status info is not in the Set, but in the normal StatusInfo.
            var info = response.statusInfoSet ?? new[] {response.statusInfo};
            var list = new List<User>();

            for (var i = 0; i < info.Length; i++)
            {
                if (info[i].codeMajor == StatusInfoDTypeCodeMajor.success)
                {
                    list.Add(personList[i].ToItsUser(syncKeyList[i]));
                }
            }

            return list;
        }


        /// <summary>
        /// Creates an ImsEs person based on the data of an Itslearning user.
        /// </summary>
        /// <param name="itsUser">The Itslearning user.</param>
        /// <param name="checkRequired">True/False, To show if the required fields of the user must be checked 
        /// (in case of updating it is not needed to check these fields).</param>
        /// <returns>An ImsEs person.</returns>
        /// <exception cref="ValidationException">When the given User is not valid 
        /// (not all required properties are filled in).</exception>
        public static person ToImsEsPerson(this User itsUser, bool checkRequired)
        {
            // First check if the given user is valid:
            if (checkRequired)
                itsUser.CheckRequired();

            itsUser.CheckValid();

            // Create a new person.
            var newPerson = new person();
            
            // Add the Names.
            var nameList = new List<NameDTypePartName>();

            if (!string.IsNullOrEmpty(itsUser.FirstName))
            {
                nameList.Add(new NameDTypePartName {namePartType = FirstName, namePartValue = itsUser.FirstName});
            }
            if (!string.IsNullOrEmpty(itsUser.LastName))
            {
                nameList.Add(new NameDTypePartName { namePartType = LastName, namePartValue = itsUser.LastName });
            }
            newPerson.name = new NameDType {partName = nameList.ToArray()};

            // Add the Username.
            if (itsUser.UserName != null)
            {
                newPerson.userId = new UserIdDType
                    {
                        userIdValue = itsUser.UserName,
                    };
            }

            // Add the role (enums are always set):
            newPerson.institutionRole = new[]
                {
                    new InstitutionRoleDType
                        {
                            institutionRoleType = itsUser.Profile.ToInstitutionRole(),
                            primaryRoleType = true
                        }
                };
            

            // Test if the birthdate is set, otherwise it happens that the birthdate is set to 30/12/1899.
            if (itsUser.BirthDate != DateTime.MinValue)
            {
                newPerson.demographics = new DemographicsDType
                    {
                        bday = itsUser.BirthDate,
                        bdaySpecified = true
                    };
            }

            // Address is not required, and must therefore be tested.
            if (itsUser.Address != null)
            {
                newPerson.address = new AddressDType
                    {
                        street = new[] {itsUser.Address.AddressLine1, itsUser.Address.AddressLine2},
                        postcode = itsUser.Address.PostalCode,
                        locality = itsUser.Address.City
                    };
            }

            if (!string.IsNullOrEmpty(itsUser.Email))
            {
                newPerson.email = itsUser.Email;
            }
            if (!string.IsNullOrEmpty(itsUser.Password))
            {
                newPerson.userId.passWord = itsUser.Password;
            }
            
            // Phone numbers only have to be added if they are filled in.
            var telList = new List<TelDType> ();

            if (!string.IsNullOrEmpty(itsUser.Phone))
            {
                telList.Add(new TelDType
                    {
                        telType = TelDTypeTelType.Voice,
                        telValue = itsUser.Phone,
                        telTypeSpecified = true
                    });
            }
            if (!string.IsNullOrEmpty(itsUser.CellPhone))
            {
                telList.Add(new TelDType
                    {
                        telType = TelDTypeTelType.Mobile,
                        telValue = itsUser.CellPhone,
                        telTypeSpecified = true
                    });
            }

            newPerson.tel = telList.ToArray();
            
            // The extension field has to be done seperately, as it consists of multiple parts.
            var extensionList = new List<extensionField>();

            if (itsUser.CustomStrings != null)
            {
                if (!string.IsNullOrEmpty(itsUser.CustomStrings.CustomString1))
                {
                    extensionList.Add(new extensionField
                    {
                        fieldName = Custom0,
                        fieldType = FieldString,
                        fieldValue = itsUser.CustomStrings.CustomString1
                    });
                }
                if (!string.IsNullOrEmpty(itsUser.CustomStrings.CustomString2))
                {
                    extensionList.Add(new extensionField
                    {
                        fieldName = Custom1,
                        fieldType = FieldString,
                        fieldValue = itsUser.CustomStrings.CustomString2
                    });
                }
                if (!string.IsNullOrEmpty(itsUser.CustomStrings.CustomString3))
                {
                    extensionList.Add(new extensionField
                    {
                        fieldName = Custom2,
                        fieldType = FieldString,
                        fieldValue = itsUser.CustomStrings.CustomString3
                    });
                }
                if (!string.IsNullOrEmpty(itsUser.CustomStrings.CustomString4))
                {
                    extensionList.Add(new extensionField
                    {
                        fieldName = Custom3,
                        fieldType = FieldString,
                        fieldValue = itsUser.CustomStrings.CustomString4
                    });
                }
                if (!string.IsNullOrEmpty(itsUser.CustomStrings.CustomString5))
                {
                    extensionList.Add(new extensionField
                    {
                        fieldName = Custom4,
                        fieldType = FieldString,
                        fieldValue = itsUser.CustomStrings.CustomString5
                    });
                }
            }

            if (itsUser.EmailServer != null)
            {
                if (!string.IsNullOrEmpty(itsUser.EmailServer.Email))
                {
                    extensionList.Add(new extensionField
                    {
                        fieldName = EmailServerEmail,
                        fieldType = FieldString,
                        fieldValue = itsUser.EmailServer.Email
                    });
                }
                if (!string.IsNullOrEmpty(itsUser.EmailServer.Server))
                {
                    extensionList.Add(new extensionField
                    {
                        fieldName = EmailServerServer,
                        fieldType = FieldString,
                        fieldValue = itsUser.EmailServer.Server
                    });
                }
                if (!string.IsNullOrEmpty(itsUser.EmailServer.Port))
                {
                    extensionList.Add(new extensionField
                        {
                            fieldName = EmailServerPort,
                            fieldType = FieldString,
                            fieldValue = itsUser.EmailServer.Port
                        });
                }
                if (!string.IsNullOrEmpty(itsUser.EmailServer.Type))
                {
                    extensionList.Add(new extensionField
                        {
                            fieldName = EmailServerType,
                            fieldType = FieldString,
                            fieldValue = itsUser.EmailServer.Type
                        });
                }
                if (!string.IsNullOrEmpty(itsUser.EmailServer.UserId))
                {
                    extensionList.Add(new extensionField
                        {
                            fieldName = EmailServerUserid,
                            fieldType = FieldString,
                            fieldValue = itsUser.EmailServer.UserId
                        });
                }
                if (!string.IsNullOrEmpty(itsUser.EmailServer.Password))
                {
                    extensionList.Add(new extensionField
                        {
                            fieldName = EmailServerPassword,
                            fieldType = FieldString,
                            fieldValue = itsUser.EmailServer.Password
                        });
                }
            }

            if (itsUser.MetaData != null)
            {
                extensionList.AddRange(
                    itsUser.MetaData.Where(x => !string.IsNullOrEmpty(x.Value))
                           .Select(x => new extensionField
                               {
                                   fieldName = (x.Type.StartsWith(Meta)) ? x.Type : Meta + x.Type,
                                   fieldType = FieldString,
                                   fieldValue = x.Value
                               }));
            }
            
            var relationList = new List<RelationshipDType>();

            if (itsUser.Children != null && itsUser.Children.Count > 0)
            {
                relationList.AddRange(itsUser.Children.Select(child => new RelationshipDType
                    {
                        sourceId = new RelationshipDTypeSourceId {identifier = child},
                        relation = RelationshipDTypeRelation.Child
                    }));
            }

            newPerson.extension = new extensionType
                {
                    extensionField = extensionList.ToArray(),
                    relationship = relationList.ToArray()
                };

            return newPerson;
        }
        
        #endregion


        #region Hierarchy

        /// <summary>
        /// Creates a new Itslearning Hierarchy based on the ImsEs Group.
        /// </summary>
        /// <param name="imsEsGroup">The ImsEs group.</param>
        /// <param name="syncId">The new SyncId of the Hierarchy.</param>
        /// <returns>An Itslearning Hierarchy.</returns>
        public static Hierarchy ToItsHierarchy(this group imsEsGroup, string syncId)
        {
            // In case the read failed, return null.
            if (imsEsGroup == null) { return null; }

            // Also check if the group is actually a Hierarchy:
            var level = imsEsGroup.groupType.typeValue[0].level;
            if (!(level.Equals(SiteLevel) || level.Equals(SchoolLevel) || level.Equals(HierarchyLevel)))
            {
                var result = new OperationResult(ResultStatus.Error, syncId,
                                                 "Retrieved ImsEs group was not of type Hiearchy.");
                throw new ImsEsException(result);
            }

            var newHierarchy = new Hierarchy(syncId);

            if (imsEsGroup.description != null)
            {
                newHierarchy.Title = imsEsGroup.description.descShort;
                newHierarchy.Description = imsEsGroup.description.descLong;
            }

            if (imsEsGroup.relationship.Length > 1)
            {
                // Find the Parent relationship.
                foreach (var link in imsEsGroup.relationship.Where(link => link.relation == relation.Parent))
                {
                    newHierarchy.ParentHierarchyId = link.sourceId.identifier;
                    break;
                }
            }
            else if (imsEsGroup.relationship.Length == 1)
            {
                // It is the first one.
                newHierarchy.ParentHierarchyId = imsEsGroup.relationship[0].sourceId.identifier;
            }

            if (imsEsGroup.groupType != null && imsEsGroup.groupType.typeValue != null && imsEsGroup.groupType.typeValue.Length > 0)
            {
                newHierarchy.GroupType = imsEsGroup.groupType.typeValue[0].level.ToGroupType();
            }
            else
            {
                newHierarchy.GroupType = GroupType.Hierarchy;
            }
            
            if (imsEsGroup.extension != null && imsEsGroup.extension.Length > 0)
            {
                var newCourse = new Course(syncId);

                foreach (var extension in imsEsGroup.extension)
                {
                    var extensionField = extension.fieldName.ToCourseExtType();

                    switch (extensionField)
                    {
                        case CourseExtFieldType.Visibleinsearch:
                            newHierarchy.VisibleInSearch = (extension.fieldValue.ToBoolType() == BoolFieldType.True);
                            break;
                        case CourseExtFieldType.Course:
                            newCourse.Title = extension.fieldValue;
                            break;
                        case CourseExtFieldType.CourseCode:
                            newCourse.Code = extension.fieldValue;
                            break;
                        case CourseExtFieldType.CourseCredits:
                            newCourse.Credit = extension.fieldValue;
                            break;
                        case CourseExtFieldType.LegalId:
                            newHierarchy.LegalId = extension.fieldValue;
                            break;
                    }

                }
                if (newCourse.Title != null)
                {
                    newCourse.LinkedHierarchy = newHierarchy;
                    newHierarchy.AttachedCourse = newCourse;
                }
            }
            
            return newHierarchy;
        }


        /// <summary>
        /// Creates a Hierarchy list based on a GroupPair array.
        /// </summary>
        /// <param name="groupList">The GroupPair array.</param>
        /// <returns>An Hierarchy List.</returns>
        public static List<Hierarchy> ToItsHierarchyList(this GroupIdPairDType[] groupList)
        {
            return groupList.Select(x => x.group.ToItsHierarchy(x.sourcedId.identifier)).ToList();
        }
        

        /// <summary>
        /// Creates a Hierarchy list based on a Group array.
        /// </summary>
        /// <param name="groupList">The array with groups.</param>
        /// <param name="syncIdList">The array with SyncIds.</param>
        /// <param name="response">Statusinfo to check if all groups are actually retrieved.</param>
        /// <returns>An Hierarchy List.</returns>
        public static List<Hierarchy> ToItsHierarchyList(this group[] groupList, List<string> syncIdList, syncResponseHeaderInfo response)
        {
            // For Batch functions, if there is only 1 object left, the status info is not in the Set, but in the normal StatusInfo.
            var info = response.statusInfoSet ?? new[] { response.statusInfo };
            var list = new List<Hierarchy>();

            for (var i = 0; i < info.Length; i++)
            {
                if (info[i].codeMajor != StatusInfoDTypeCodeMajor.success) 
                    continue;

                var level = groupList[i].groupType.typeValue[0].level;
                if ((level.Equals(SiteLevel) || level.Equals(SchoolLevel) || level.Equals(HierarchyLevel)))
                {
                    list.Add(groupList[i].ToItsHierarchy(syncIdList[i]));
                }
            }
            
            return list;
        }


        /// <summary>
        /// Converts an Itslearning Hierarchy to an ImsEs Group.
        /// </summary>
        /// <param name="itsHierarchy">The Itslearning Hierarchy.</param>
        /// <param name="checkRequired">True/False, To show if the required fields of the hierarchy must be checked 
        /// (in case of updating it is not needed to check these fields).</param>
        /// <returns>The ImsEs Group.</returns>
        /// <exception cref="ValidationException">When the given Hierarchy is not valid (not all required properties are filled in).</exception>
        public static group ToImsEsGroup(this Hierarchy itsHierarchy, bool checkRequired)
        {
            // First check if the given Hierarchy is valid.
            if (checkRequired)
                itsHierarchy.CheckRequired();
            
            itsHierarchy.CheckValid();

            var extensionFields = new List<extensionField>
            {
                new extensionField
                {
                    fieldName = VisibleSearch,
                    fieldType = FieldBoolean,
                    fieldValue = itsHierarchy.VisibleInSearch ? BoolTrue : BoolFalse
                }
            };

            if (!string.IsNullOrWhiteSpace(itsHierarchy.LegalId))
            {
                extensionFields.Add(new extensionField
                {
                    fieldName = LegalId,
                    fieldType = FieldString,
                    fieldValue = itsHierarchy.LegalId
                });
            }

            var newGroup = new group
                {
                    groupType = new GroupTypeDType
                        {
                            scheme = Scheme, 
                            typeValue = new [] { new TypeValueDType{ level = ((int)itsHierarchy.GroupType).ToString("G") } }
                        },

                    extension = extensionFields.ToArray()
            };

            if (itsHierarchy.ParentHierarchyId != null)
            {
                newGroup.relationship = new[]
                    {
                        new RelationshipDType1
                            {
                                relation = relation.Parent,
                                sourceId = new RelationshipDTypeSourceId1 {identifier = itsHierarchy.ParentHierarchyId}
                            }
                    };
            }

            if (itsHierarchy.Title != null || itsHierarchy.Description != null)
            {
                newGroup.description = new DescriptionDType();

                if (itsHierarchy.Title != null)
                {
                    newGroup.description.descShort = itsHierarchy.Title;
                }
                if (itsHierarchy.Description != null)
                {
                    newGroup.description.descLong = itsHierarchy.Description;
                }
            }
            
            return newGroup;
        }

        #endregion


        #region Course

        /// <summary>
        /// Creates a new Itslearning Course based on the ImsEs Group.
        /// </summary>
        /// <param name="imsEsGroup">The ImsEs group.</param>
        /// <param name="syncId">The new SyncId of the Course.</param>
        /// <returns>An Itslearning Course.</returns>
        public static Course ToItsCourse(this group imsEsGroup, string syncId)
        {
            // In case the read failed, return null.
            if (imsEsGroup == null) { return null; }

            // Also check if it is actually a Course:
            var level = imsEsGroup.groupType.typeValue[0].level;
            if (!level.Equals(CourseLevel))
            {
                var result = new OperationResult(ResultStatus.Error, syncId,
                                                 "The retrieved ImsEs group is not of type Course!");
                throw new ImsEsException(result);
            }

            var newCourse = new Course(syncId);

            var newLinkedHierarchy = new Hierarchy(syncId)
                {
                    Title = imsEsGroup.description.descShort,
                    Description = imsEsGroup.description.descLong,
                };

            if (imsEsGroup.relationship.Length > 1)
            {
                // Find the Parent relationship.
                foreach (var link in imsEsGroup.relationship.Where(link => link.relation == relation.Parent))
                {
                    newLinkedHierarchy.ParentHierarchyId = link.sourceId.identifier;
                    break;
                }
            }
            else if (imsEsGroup.relationship.Length == 1)
            {
                newLinkedHierarchy.ParentHierarchyId = imsEsGroup.relationship[0].sourceId.identifier;
            }

            if (imsEsGroup.groupType != null && imsEsGroup.groupType.typeValue != null && imsEsGroup.groupType.typeValue.Length > 0)
            {
                newLinkedHierarchy.GroupType = imsEsGroup.groupType.typeValue[0].level.ToGroupType();
            }
            else
            {
                newLinkedHierarchy.GroupType = GroupType.Hierarchy;
            }

            // Walk through the extension fields, for the rest of the values.
            foreach (var extension in imsEsGroup.extension)
            {
                var extensionField = extension.fieldName.ToCourseExtType();

                switch (extensionField)
                {
                    case CourseExtFieldType.Visibleinsearch:
                        newLinkedHierarchy.VisibleInSearch = (extension.fieldValue.ToBoolType() == BoolFieldType.True);
                        break;
                    case CourseExtFieldType.Course:
                        newCourse.Title = extension.fieldValue;
                        break;
                    case CourseExtFieldType.CourseCode:
                        newCourse.Code = extension.fieldValue;
                        break;
                    case CourseExtFieldType.CourseCredits:
                        newCourse.Credit = extension.fieldValue;
                        break;
                }
            }

            // Make a connection both ways:
            newCourse.LinkedHierarchy = newLinkedHierarchy;
            newLinkedHierarchy.AttachedCourse = newCourse;

            return newCourse;
        }


        /// <summary>
        /// Creates a Course list based on a GroupPair array, throwing away the non-Course Hierarchies.
        /// </summary>
        /// <param name="groupList">The GroupPair array.</param>
        /// <returns>A Course List.</returns>
        public static List<Course> ToItsCourseList(this GroupIdPairDType[] groupList)
        {
            // Only return the Courses (based on the GroupType).
            return (from pair in groupList 
                    where pair.@group.groupType.typeValue[0].level.Equals(CourseLevel) 
                    select pair.@group.ToItsCourse(pair.sourcedId.identifier))
                    .ToList();
        }


        /// <summary>
        /// Creates a Course list based on a Group array, throwing away the non-Course Hierarchies.
        /// </summary>
        /// <param name="groupList">The Group array</param>
        /// <param name="syncIdList">The array with SyncIds.</param>
        /// <param name="response">Statusinfo to check if all groups are actually retrieved.</param>
        /// <returns>A Course List.</returns>
        public static List<Course> ToItsCourseList(this group[] groupList, List<string> syncIdList, syncResponseHeaderInfo response)
        {
            // For Batch functions, if there is only 1 object left, the status info is not in the Set, but in the normal StatusInfo.
            var info = response.statusInfoSet ?? new[] { response.statusInfo };
            var list = new List<Course>();

            for (var i = 0; i < info.Length; i++)
            {
                if (info[i].codeMajor != StatusInfoDTypeCodeMajor.success) 
                    continue;

                var level = groupList[i].groupType.typeValue[0].level;
                if (level.Equals(CourseLevel))
                {
                    list.Add(groupList[i].ToItsCourse(syncIdList[i]));
                }
            }

            return list;
        }


        /// <summary>
        /// Converts an Itslearning Course to an ImsEs Group.
        /// </summary>
        /// <param name="itsCourse">The Itslearning Course.</param>
        /// <param name="checkRequired">True/False, To show if the required fields of the course must be checked 
        /// (in case of updating it is not needed to check these fields).</param>
        /// <returns>The ImsEs Group.</returns>
        /// <exception cref="ValidationException">When the given Course is not valid (not all required properties are filled in).</exception>
        public static group ToImsEsGroup(this Course itsCourse, bool checkRequired)
        {
            // First check if the given Hierarchy is valid.
            if (checkRequired)
                itsCourse.CheckRequired();

            itsCourse.CheckValid();

            var newGroup = new group
            {
                groupType = new GroupTypeDType
                {
                    scheme = Scheme,
                    typeValue = new[] { new TypeValueDType { level = ((int)itsCourse.LinkedHierarchy.GroupType).ToString("G") } }
                },
            };

            var extList = new List<extensionField>();

            if (itsCourse.LinkedHierarchy != null)
            {
                var itsHierarchy = itsCourse.LinkedHierarchy;
                if (itsHierarchy.Title != null || itsHierarchy.Description != null)
                {
                    newGroup.description = new DescriptionDType();

                    if (itsHierarchy.Title != null)
                    {
                        newGroup.description.descShort = itsHierarchy.Title;
                    }
                    if (itsHierarchy.Description != null)
                    {
                        newGroup.description.descLong = itsHierarchy.Description;
                    }
                }
                if (itsHierarchy.ParentHierarchyId != null)
                {
                    newGroup.relationship = new[]
                        {
                            new RelationshipDType1
                                {
                                    relation = relation.Parent,
                                    sourceId = new RelationshipDTypeSourceId1 { identifier = itsHierarchy.ParentHierarchyId }
                                }
                        };
                }

                // Visible in Search is always set.
                extList.Add(new extensionField
                    {
                        fieldName = VisibleSearch,
                        fieldType = FieldBoolean,
                        fieldValue = itsCourse.LinkedHierarchy.VisibleInSearch ? BoolTrue : BoolFalse
                    });
            }

            if (itsCourse.Title != null)
            {
                extList.Add(new extensionField
                {
                    fieldName = Course,
                    fieldType = FieldString,
                    fieldValue = itsCourse.Title
                });
            }
            if (itsCourse.Code != null)
            {
                extList.Add(new extensionField
                            {
                                fieldName = CourseCode,
                                fieldType = FieldString,
                                fieldValue = itsCourse.Code
                            });
            }
            if (itsCourse.Credit != null)
            {
                extList.Add(new extensionField
                            {
                                fieldName = CourseCredits,
                                fieldType = FieldString,
                                fieldValue = itsCourse.Credit
                            });
            }
            newGroup.extension = extList.ToArray();
            
            return newGroup;
        }

        #endregion


        #region CourseGroup

        /// <summary>
        /// Creates a new Itslearning CourseGroup based on the ImsEs Group.
        /// </summary>
        /// <param name="imsEsGroup">The ImsEs group.</param>
        /// <param name="syncId">The new SyncId of the CourseGroup.</param>
        /// <returns>An Itslearning CourseGroup.</returns>
        public static CourseGroup ToItsCourseGroup(this group imsEsGroup, string syncId)
        {
            // In case the read failed, return null.
            if (imsEsGroup == null) { return null; }

            // Also check if it is actually a CourseGroup.
            var level = imsEsGroup.groupType.typeValue[0].level;
            if (!level.Equals(CourseGroupLevel))
            {
                var result = new OperationResult(ResultStatus.Error, syncId,
                                                 "The retrieved ImsEs group is not of type Course!");
                throw new ImsEsException(result);
            }

            var newCourseGroup = new CourseGroup(syncId);

            if (imsEsGroup.description != null)
            {
                newCourseGroup.Title = imsEsGroup.description.descShort;
                newCourseGroup.Description = imsEsGroup.description.descLong;
            }

            if (imsEsGroup.relationship.Length > 1)
            {
                // Find the Parent relationship.
                foreach (var link in imsEsGroup.relationship.Where(link => link.relation == relation.Parent))
                {
                    newCourseGroup.ParentHierarchyId = link.sourceId.identifier;
                    break;
                }
            }
            else if (imsEsGroup.relationship.Length == 1)
            {
                // It is the first one.
                newCourseGroup.ParentHierarchyId = imsEsGroup.relationship[0].sourceId.identifier;
            }

            // It is always a CourseGroup:
            newCourseGroup.GroupType = GroupType.CourseGroup;

            if (imsEsGroup.extension != null && imsEsGroup.extension.Length > 0)
            {
                foreach (var extension in imsEsGroup.extension)
                {
                    var extensionField = extension.fieldName.ToCourseExtType();

                    switch (extensionField)
                    {
                        case CourseExtFieldType.Visibleinsearch:
                            newCourseGroup.VisibleInSearch = (extension.fieldValue.ToBoolType() == BoolFieldType.True);
                            break;
                    }
                }
            }

            return newCourseGroup;
        }

        /// <summary>
        /// Creates a CourseGroup list based on a GroupPair array.
        /// </summary>
        /// <param name="groupList">The GroupPair array.</param>
        /// <returns>An CourseGroup List.</returns>
        public static List<CourseGroup> ToItsCourseGroupList(this GroupIdPairDType[] groupList)
        {
            return groupList.Select(x => x.group.ToItsCourseGroup(x.sourcedId.identifier)).ToList();
        }


        /// <summary>
        /// Creates a CourseGroup list based on a Group array.
        /// </summary>
        /// <param name="groupList">The array with groups.</param>
        /// <param name="syncIdList">The array with SyncIds.</param>
        /// <param name="response">Statusinfo to check if all groups are actually retrieved.</param>
        /// <returns>An CourseGroup List.</returns>
        public static List<CourseGroup> ToItsCourseGroupList(this group[] groupList, List<string> syncIdList, syncResponseHeaderInfo response)
        {
            // For Batch functions, if there is only 1 object left, the status info is not in the Set, but in the normal StatusInfo.
            var info = response.statusInfoSet ?? new[] { response.statusInfo };
            var list = new List<CourseGroup>();

            for (var i = 0; i < info.Length; i++)
            {
                if (info[i].codeMajor != StatusInfoDTypeCodeMajor.success) 
                    continue;

                var level = groupList[i].groupType.typeValue[0].level;
                if (level.Equals(CourseGroupLevel))
                {
                    list.Add(groupList[i].ToItsCourseGroup(syncIdList[i]));
                }
            }

            return list;
        }


        /// <summary>
        /// Converts an Itslearning CourseGroup to an ImsEs Group.
        /// </summary>
        /// <param name="itsCourseGroup">The Itslearning CourseGroup.</param>
        /// <param name="checkRequired">True/False, To show if the required fields of the CourseGroup must be checked 
        /// (in case of updating it is not needed to check these fields).</param>
        /// <returns>The ImsEs Group.</returns>
        /// <exception cref="ValidationException">When the given CourseGroup is not valid (not all required properties are filled in).</exception>
        public static group ToImsEsGroup(this CourseGroup itsCourseGroup, bool checkRequired)
        {
            // First check if the given CourseGroup is valid.
            if (checkRequired)
                itsCourseGroup.CheckRequired();

            itsCourseGroup.CheckValid();

            var newGroup = new group
            {
                groupType = new GroupTypeDType
                {
                    scheme = Scheme,
                    typeValue = new[] { new TypeValueDType { level = ((int)itsCourseGroup.GroupType).ToString("G") } }
                },

                extension = new[]
                        {
                            new extensionField
                                {
                                    fieldName = VisibleSearch,
                                    fieldType = FieldBoolean,
                                    fieldValue = itsCourseGroup.VisibleInSearch ? BoolTrue : BoolFalse
                                }
                        }
            };

            if (itsCourseGroup.ParentHierarchyId != null)
            {
                newGroup.relationship = new[]
                    {
                        new RelationshipDType1
                            {
                                relation = relation.Parent,
                                sourceId =
                                    new RelationshipDTypeSourceId1 {identifier = itsCourseGroup.ParentHierarchyId}
                            }
                    };
            }

            if (itsCourseGroup.Title != null || itsCourseGroup.Description != null)
            {
                newGroup.description = new DescriptionDType();

                if (itsCourseGroup.Title != null)
                {
                    newGroup.description.descShort = itsCourseGroup.Title;
                }
                if (itsCourseGroup.Description != null)
                {
                    newGroup.description.descLong = itsCourseGroup.Description;
                }
            }

            return newGroup;
        }

        #endregion


        #region Membership

        /// <summary>
        /// Converts a MembershipPair to an ImsEs Membership.
        /// </summary>
        /// <param name="membership">The MembershipPair.</param>
        /// <returns>An ImsEs membership.</returns>
        public static membership ToImsEsMembership(this MembershipPair membership)
        {
            return ToImsEsMembership(membership.UserSyncKey, membership.HierarchySyncId, membership.Role);
        }

        /// <summary>
        /// Converts a MembershipPair to an ImsEs Membership.
        /// </summary>
        /// <param name="userSyncKey">The User.</param>
        /// <param name="hierarchySyncId">The Hierarchy.</param>
        /// <param name="role">Role of the User.</param>
        /// <returns>An ImsEs membership.</returns>
        public static membership ToImsEsMembership(string userSyncKey, string hierarchySyncId, HierarchyRole role)
        {
            return new membership
            {
                groupSourcedId = new membershipGroupSourcedId { identifier = hierarchySyncId },
                member = new[]
                        {
                            new MemberDType
                                {
                                    memberSourcedId = new MemberDTypeMemberSourcedId {identifier = userSyncKey},
                                    role = new[] {new RoleDType {roleType = role.ToImsEsRole()}}
                                }
                        }


            };
        }


        /// <summary>
        /// Converts a MembershipPair to an ImsEs Membership.
        /// </summary>
        /// <param name="membership">The MembershipPair.</param>
        /// <returns>An ImsEs membership.</returns>
        public static membership ToImsEsHomeMembership(this MembershipPair membership)
        {
            return ToImsEsHomeMembership(membership.UserSyncKey, membership.HierarchySyncId);
        }

        /// <summary>
        /// Converts a MembershipPair to an ImsEs Membership.
        /// </summary>
        /// <param name="userSyncKey">The User.</param>
        /// <param name="hierarchySyncId">The Hierarchy.</param>
        /// <returns>An ImsEs membership.</returns>
        public static membership ToImsEsHomeMembership(string userSyncKey, string hierarchySyncId)
        {
            var extension = new[] { new extensionField
                        {
                            fieldName = HomeOrganisation,
                            fieldType = FieldBoolean,
                            fieldValue = "1"
                        }
                };

            return new membership
                {
                    groupSourcedId = new membershipGroupSourcedId {identifier = hierarchySyncId},
                    member = new[]
                        {
                            new MemberDType
                                {
                                    memberSourcedId = new MemberDTypeMemberSourcedId {identifier = userSyncKey},
                                    role = new[] {new RoleDType {extension = extension}}
                                }
                        }


                };
        }

        #endregion

    }
}
